// Boot stub, loads code from 0x1000 in ROM into RSP DMEM and runs it.

//define FILL_ALL_DMEM()

arch n64.cpu
endian msb

// CP0
constant Count(9)
constant Compare(11)
constant Cause(16)
constant TagLo(28)

constant r0(0)
constant t0(8)
constant t1(9)
constant t2(10)
constant t3(11)

constant ICACHE_LINE(32)
constant ICACHE_SIZE(0x4000) // 32K
constant ICACHE_INDEX_STORE_TAG((2<<2)|0)
constant ICACHE_FILL((5<<2)|0)
constant LOAD_ADDR(0xb000'1000)
constant SP_DMEM(0xa400'0000)
constant DMEM_SIZE(0x1000)

macro align(evaluate alignment) {
  evaluate overhang(pc() % {alignment})
  if {overhang} {
    fill {alignment} - {overhang}
  }
}

origin 0
base SP_DMEM

// Initial ROM timing info
  db 0x80, 0x37, 0x12, 0x40

// Initial CPU timing
  dw 15

// Pad out to IPL3 entrypoint
align(0x40)

if pc() != 0xa400'0040 {
  error "entrypoint is wrong"
}
Entrypoint:
  mtc0 r0, Cause
// This will prevent the timer interrupt from hitting until the counter wraps around
  mtc0 r0, Count
  mtc0 r0, Compare

if {defined FILL_ALL_DMEM} {
// Initialize instruction cache
  mtc0 r0, TagLo
  lli t0, 0
  addi t1, t0, ICACHE_SIZE-ICACHE_LINE // 16K
-
  cache ICACHE_INDEX_STORE_TAG, 0(t0)
  bne t0, t1,-
  addi t0, ICACHE_LINE

// Begin using cache
  la t0, Cached
  jr t0
} else {
  j Copier
}
  lui t1, SP_DMEM>>16

if {defined FILL_ALL_DMEM} {
  // Cached addresses here out
  base pc() - 0x2000'0000

  align(ICACHE_LINE)
Cached:
  // Fill second icache line before we overwrite DMEM
    cache ICACHE_FILL, ICACHE_LINE*1(t0)
} else {
  fill SP_DMEM + DMEM_SIZE - 0x40 - pc() 
Copier:
}

// Load the data into DMEM
  la t0, LOAD_ADDR
if {defined FILL_ALL_DMEM} {
  addi t2, t0, DMEM_SIZE
} else {
  addi t2, t0, Copier - SP_DMEM
}
-
  lw t3, 0(t0)
  addi t0, 4
  sw t3, 0(t1)
  bne t0, t2,-
  addi t1, 4

// Jump, uncached, into the newly copied data in DMEM
  lui t1, SP_DMEM>>16
  jr t1
  nop

if {defined FILL_ALL_DMEM} {
  if pc() - Cached > ICACHE_LINE*2 {
    error "max 2 icache lines expected"
  }
} else {
  if pc() > SP_DMEM + DMEM_SIZE - 8 {
    error "copier too large"
  }
}

fill DMEM_SIZE - 8 - origin()
// Collide with expected hash for 6102
dw 0, 0
